@@ -15,49 +15,58 @@ enum class CalcDyDx {
 namespace glmmr {
 
 class calculator {
-  public:
-    std::vector<Do> instructions;
-    intvec indexes;
-    dblvec y;
-    std::array<double,10> numbers;
-    strvec parameter_names;
-    strvec data_names;
-    ArrayXd variance = ArrayXd::Constant(1,1.0);
-    int data_count = 0;
-    int parameter_count = 0;
-    int user_number_count = 0;
-    bool any_nonlinear = false;
-    calculator() {};
-    
-    template<CalcDyDx dydx>
-    dblvec calculate(const int i, 
-                     const dblvec& parameters, 
-                     const MatrixXd& data,
-                     const int j = 0,
-                     const int parameterIndex = 0,
-                     const double extraData = 0.0,
-                     const int n = 0) const;
-    
-    calculator& operator= (const glmmr::calculator& calc);
-    VectorXd linear_predictor(const dblvec& parameters,const MatrixXd& data);
-    MatrixXd jacobian(const dblvec& parameters, const MatrixXd& data);
-    MatrixXd jacobian(const dblvec& parameters,const MatrixXd& data,const VectorXd& extraData);
-    MatrixXd jacobian(const dblvec& parameters,const MatrixXd& data,const MatrixXd& extraData);
-    matrix_matrix jacobian_and_hessian(const dblvec& parameters,const MatrixXd& data,const MatrixXd& extraData);
-    vector_matrix jacobian_and_hessian(const dblvec& parameters);
-    void print_instructions() const;
-    void print_names(bool data = true, bool parameters = false) const;
+public:
+  std::vector<Do>       instructions; // vector of insructions to execute
+  intvec                indexes; // indexes of data or parameter vectors
+  dblvec                y;  // outcome data
+  std::array<double,20> numbers = {0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0};
+  strvec                parameter_names; //vector of parameter names
+  strvec                data_names; // vector of data names
+  ArrayXd               variance = ArrayXd::Constant(1,1.0); // variance values
+  int                   data_count = 0; // number of data items
+  int                   parameter_count = 0; // number of parameters
+  int                   user_number_count = 0; // number of numbers in the function
+  int                   data_size = 0; // number of data items in the calculation
+  bool                  any_nonlinear = false; // for linear predictor - any non-linear functions?
+  MatrixXd              data = MatrixXd::Zero(1,1); // the data for the calculation
+  dblvec                parameters;
+  intvec                parameter_indexes;
+  calculator() {};
+  
+  template<CalcDyDx dydx>
+  dblvec calculate(const int i, 
+                   const int j = 0,
+                   const int parameterIndex = 0,
+                   const double extraData = 0.0) const;
+  
+  calculator& operator= (const glmmr::calculator& calc);
+  VectorXd      linear_predictor();
+  MatrixXd      jacobian();
+  MatrixXd      jacobian(const VectorXd& extraData);
+  MatrixXd      jacobian(const MatrixXd& extraData);
+  MatrixMatrix  jacobian_and_hessian(const MatrixXd& extraData);
+  VectorMatrix  jacobian_and_hessian();
+  void          update_parameters(const dblvec& parameters_in);
+  double        get_covariance_data(const int i, const int j, const int fn);
+  void          print_instructions() const;
+  void          print_names(bool data = true, bool parameters = false) const;
 };
 
 }
 
-inline VectorXd glmmr::calculator::linear_predictor(const dblvec& parameters, 
-                                                    const MatrixXd& data){
+inline void glmmr::calculator::update_parameters(const dblvec& parameters_in){
+#ifdef R_BUILD
+  if(parameters_in.size() < parameter_count)Rcpp::stop("Expecting "+std::to_string(parameter_count)+" parameters in calculator but got "+std::to_string(parameters_in.size()));
+#endif
+  for(int i = 0; i < parameter_indexes.size(); i++)parameters[i] = parameters_in[parameter_indexes[i]];
+}
+
+inline VectorXd glmmr::calculator::linear_predictor(){
   int n = data.rows();
   VectorXd x(n);
 #pragma omp parallel for
   for(int i = 0; i < n; i++){
-    x(i) = calculate<CalcDyDx::None>(i,parameters,data)[0];
+    x(i) = calculate<CalcDyDx::None>(i)[0];
   }
   return x;
 };
@@ -72,19 +81,22 @@ inline glmmr::calculator& glmmr::calculator::operator= (const glmmr::calculator&
   data_count = calc.data_count;
   parameter_count = calc.parameter_count;
   any_nonlinear = calc.any_nonlinear;
+  data.conservativeResize(calc.data.rows(),calc.data.cols());
+  data = calc.data;
+  parameters.resize(calc.parameters.size());
+  parameters = calc.parameters;
   return *this;
 };
 
-inline MatrixXd glmmr::calculator::jacobian(const dblvec& parameters,
-                                            const MatrixXd& data){
+inline MatrixXd glmmr::calculator::jacobian(){
   int n = data.rows();
-  #if defined(ENABLE_DEBUG) && defined(R_BUILD)
+#if defined(ENABLE_DEBUG) && defined(R_BUILD)
   if(n==0)Rcpp::stop("No data initialised in calculator");
-  #endif
+#endif
   MatrixXd J(n,parameter_count);
 #pragma omp parallel for
   for(int i = 0; i<n ; i++){
-    dblvec out = calculate<CalcDyDx::BetaFirst>(i,parameters,data);
+    dblvec out = calculate<CalcDyDx::BetaFirst>(i);
     for(int j = 0; j<parameter_count; j++){
       J(i,j) = out[j+1];
     }
@@ -92,17 +104,15 @@ inline MatrixXd glmmr::calculator::jacobian(const dblvec& parameters,
   return J;
 };
 
-inline MatrixXd glmmr::calculator::jacobian(const dblvec& parameters,
-                                            const MatrixXd& data,
-                                            const VectorXd& extraData){
+inline MatrixXd glmmr::calculator::jacobian(const VectorXd& extraData){
   int n = data.rows();
-  #if defined(ENABLE_DEBUG) && defined(R_BUILD)
+#if defined(ENABLE_DEBUG) && defined(R_BUILD)
   if(n==0)Rcpp::stop("No data initialised in calculator");
-  #endif 
+#endif 
   MatrixXd J(n,parameter_count);
 #pragma omp parallel for
   for(int i = 0; i<n ; i++){
-    dblvec out = calculate<CalcDyDx::BetaFirst>(i,parameters,data,0,0,extraData(i));
+    dblvec out = calculate<CalcDyDx::BetaFirst>(i,0,0,extraData(i));
     for(int j = 0; j<parameter_count; j++){
       J(i,j) = out[j+1];
     }
@@ -110,15 +120,13 @@ inline MatrixXd glmmr::calculator::jacobian(const dblvec& parameters,
   return J;
 };
 
-inline MatrixXd glmmr::calculator::jacobian(const dblvec& parameters,
-                                            const MatrixXd& data,
-                                            const MatrixXd& extraData){
+inline MatrixXd glmmr::calculator::jacobian(const MatrixXd& extraData){
   int n = data.rows();
   
-  #if defined(ENABLE_DEBUG) && defined(R_BUILD)
+#if defined(ENABLE_DEBUG) && defined(R_BUILD)
   if(n==0)Rcpp::stop("No data initialised in calculator");
   if(extraData.rows()!=n)Rcpp::stop("Extra data not of length n");
-  #endif
+#endif
   
   int iter = extraData.cols();
   MatrixXd J = MatrixXd::Zero(parameter_count,n);
@@ -126,7 +134,7 @@ inline MatrixXd glmmr::calculator::jacobian(const dblvec& parameters,
   for(int i = 0; i<n ; i++){
     dblvec out;
     for(int k = 0; k < iter; k++){
-      out = calculate<CalcDyDx::BetaFirst>(i,parameters,data,0,0,extraData(i,k));
+      out = calculate<CalcDyDx::BetaFirst>(i,0,0,extraData(i,k));
       for(int j = 0; j < parameter_count; j++){
         J(j,i) += out[1+j]/iter;
       }
@@ -136,16 +144,14 @@ inline MatrixXd glmmr::calculator::jacobian(const dblvec& parameters,
 };
 
 
-inline matrix_matrix glmmr::calculator::jacobian_and_hessian(const dblvec& parameters,
-                                                             const MatrixXd& data,
-                                                             const MatrixXd& extraData){
+inline MatrixMatrix glmmr::calculator::jacobian_and_hessian(const MatrixXd& extraData){
   int n = data.rows();
-  matrix_matrix result(parameter_count,parameter_count,parameter_count,n);
+  MatrixMatrix result(parameter_count,parameter_count,parameter_count,n);
   
-  #ifdef ENABLE_DEBUG
+#ifdef ENABLE_DEBUG
   if(n==0)Rcpp::stop("No data initialised in calculator");
   if(extraData.rows()!=n)Rcpp::stop("Extra data not of length n");
-  #endif
+#endif
   
   int iter = extraData.cols();
   int n2d = parameter_count*(parameter_count + 1)/2;
@@ -154,7 +160,7 @@ inline matrix_matrix glmmr::calculator::jacobian_and_hessian(const dblvec& param
 #pragma omp parallel for collapse(2)
   for(int i = 0; i<n ; i++){
     for(int k = 0; k < iter; k++){
-      dblvec out = calculate<CalcDyDx::BetaSecond>(i,parameters,data,0,0,extraData(i,k));
+      dblvec out = calculate<CalcDyDx::BetaSecond>(i,0,0,extraData(i,k));
       for(int j = 0; j < parameter_count; j++){
         J(j,i) += out[1+j]/iter;
       }
@@ -178,13 +184,13 @@ inline matrix_matrix glmmr::calculator::jacobian_and_hessian(const dblvec& param
   return result;
 };
 
-inline vector_matrix glmmr::calculator::jacobian_and_hessian(const dblvec& parameters){
-  vector_matrix result(parameter_count);
+inline VectorMatrix glmmr::calculator::jacobian_and_hessian(){
+  VectorMatrix result(parameter_count);
   int n2d = parameter_count*(parameter_count + 1)/2;
   VectorXd H = VectorXd::Zero(n2d);
   VectorXd J = VectorXd::Zero(parameter_count);
   MatrixXd dat = MatrixXd::Zero(1,1);
-  dblvec out = calculate<CalcDyDx::BetaSecond>(0,parameters,dat,0,2,0);
+  dblvec out = calculate<CalcDyDx::BetaSecond>(0,0,2,0);
   for(int j = 0; j < parameter_count; j++){
     J(j,0) += out[1+j];
   }
@@ -205,14 +211,19 @@ inline vector_matrix glmmr::calculator::jacobian_and_hessian(const dblvec& param
   return result;
 };
 
+inline double glmmr::calculator::get_covariance_data(const int i, const int j, const int fn){
+  int i1 = i < j ? (data_size-1)*i - ((i-1)*i/2) + (j-i-1) : (data_size-1)*j - ((j-1)*j/2) + (i-j-1);
+#if defined(ENABLE_DEBUG) && defined(R_BUILD)
+  if(i1 >= data.rows())Rcpp::stop("PushCovData: Index out of range: "+std::to_string(i1)+" versus "+std::to_string(data.rows()));
+#endif
+  return data(i1,fn);
+}
+
 template<CalcDyDx dydx>
 inline dblvec glmmr::calculator::calculate(const int i, 
-                                           const dblvec& parameters, 
-                                           const MatrixXd& data,
                                            const int j,
                                            const int parameterIndex,
-                                           const double extraData,
-                                           const int n) const {
+                                           const double extraData) const {
   int idx_iter = 0;
   double a,b;
   std::stack<double> stack;
@@ -257,246 +268,247 @@ inline dblvec glmmr::calculator::calculate(const int i,
   for(const auto& k: instructions){
     switch(k){
     case Do::PushData:
-  {
-// debugging statements to find possible errors
+    {
+      // debugging statements to find possible errors
 #if defined(ENABLE_DEBUG) && defined(R_BUILD)
-    if(idx_iter >= indexes.size())Rcpp::stop("Index out of range: case 0 idx iter: "+std::to_string(idx_iter)+" versus "+std::to_string(indexes.size()));
-    if(indexes[idx_iter] >= data.cols())Rcpp::stop("Index out of range: case 0 indexes: "+std::to_string(indexes[idx_iter])+" versus "+std::to_string(data.cols()));
-    if(i >= data.rows())Rcpp::stop("Row index out of range: case 0: "+std::to_string(i)+" versus "+std::to_string(data.rows()));
+      if(idx_iter >= indexes.size())Rcpp::stop("Index out of range: case 0 idx iter: "+std::to_string(idx_iter)+" versus "+std::to_string(indexes.size()));
+      if(indexes[idx_iter] >= data.cols())Rcpp::stop("Index out of range: case 0 indexes: "+std::to_string(indexes[idx_iter])+" versus "+std::to_string(data.cols()));
+      if(i >= data.rows())Rcpp::stop("Row index out of range: case 0: "+std::to_string(i)+" versus "+std::to_string(data.rows()));
 #endif
-    stack.push(data(i,indexes[idx_iter]));
-    if constexpr (dydx == CalcDyDx::BetaFirst || dydx == CalcDyDx::BetaSecond)addZeroDx();
-    if constexpr (dydx == CalcDyDx::BetaSecond)addZeroDx2();
-    if constexpr (dydx == CalcDyDx::XBeta){
-      if(parameterIndex == indexes[idx_iter]){
-        first_dx[0].push(1.0);
-      } else {
-        first_dx[0].push(0.0);
+      stack.push(data(i,indexes[idx_iter]));
+      if constexpr (dydx == CalcDyDx::BetaFirst || dydx == CalcDyDx::BetaSecond)addZeroDx();
+      if constexpr (dydx == CalcDyDx::BetaSecond)addZeroDx2();
+      if constexpr (dydx == CalcDyDx::XBeta){
+        if(parameterIndex == indexes[idx_iter]){
+          first_dx[0].push(1.0);
+        } else {
+          first_dx[0].push(0.0);
+        }
+        for(int i = 0; i < parameter_count; i++)first_dx[i+1].push(0.0);
+        addZeroDx2();
       }
-      for(int i = 0; i < parameter_count; i++)first_dx[i+1].push(0.0);
-      addZeroDx2();
+      if constexpr (dydx == CalcDyDx::Zu)first_dx[0].push(0.0);
+      idx_iter++;
+      break;
     }
-    if constexpr (dydx == CalcDyDx::Zu)first_dx[0].push(0.0);
-    idx_iter++;
-    break;
-  }
     case Do::PushCovData:
-  {
+    {
 #if defined(ENABLE_DEBUG) && defined(R_BUILD)
-    if(idx_iter >= indexes.size())Rcpp::stop("Index out of range: case 1 idx iter: "+std::to_string(idx_iter)+" versus "+std::to_string(indexes.size()));
-    if(indexes[idx_iter] >= data.cols())Rcpp::stop("Index out of range: case 1 indexes: "+std::to_string(indexes[idx_iter])+" versus "+std::to_string(data.cols()));
+      if(idx_iter >= indexes.size())Rcpp::stop("Index out of range: case 1 idx iter: "+std::to_string(idx_iter)+" versus "+std::to_string(indexes.size()));
+      if(indexes[idx_iter] >= data.cols())Rcpp::stop("Index out of range: case 1 indexes: "+std::to_string(indexes[idx_iter])+" versus "+std::to_string(data.cols()));
 #endif
-    
-    if(i==j){
-      stack.push(0.0);
-    } else {
-      int i1 = i < j ? (n-1)*i - ((i-1)*i/2) + (j-i-1) : (n-1)*j - ((j-1)*j/2) + (i-j-1);
+      
+      if(i==j){
+        stack.push(0.0);
+      } else {
+        int i1 = i < j ? (data_size-1)*i - ((i-1)*i/2) + (j-i-1) : (data_size-1)*j - ((j-1)*j/2) + (i-j-1);
 #if defined(ENABLE_DEBUG) && defined(R_BUILD)
-      if(i1 >= data.rows())Rcpp::stop("PushCovData: Index out of range: "+std::to_string(i1)+" versus "+std::to_string(data.rows()));
+        if(i1 >= data.rows())Rcpp::stop("PushCovData: Index out of range: "+std::to_string(i1)+" versus "+std::to_string(data.rows()));
 #endif
-      stack.push(data(i1,indexes[idx_iter]));
+        stack.push(data(i1,indexes[idx_iter]));
+      }
+      if constexpr (dydx != CalcDyDx::None)addZeroDx();
+      if constexpr (dydx == CalcDyDx::BetaSecond || dydx == CalcDyDx::XBeta)addZeroDx2();
+      idx_iter++;
+      break;
     }
-    if constexpr (dydx != CalcDyDx::None)addZeroDx();
-    if constexpr (dydx == CalcDyDx::BetaSecond || dydx == CalcDyDx::XBeta)addZeroDx2();
-    idx_iter++;
-    break;
-  }
     case Do::PushParameter:
-  {
+    {
 #if defined(ENABLE_DEBUG) && defined(R_BUILD)
-    if((unsigned)idx_iter >= indexes.size())Rcpp::stop("Index out of range: case 2 idx iter: "+std::to_string(idx_iter)+" versus "+std::to_string(indexes.size()));
-    if(indexes[idx_iter] >= parameter_count)Rcpp::stop("Index out of range: case 2 indexes: "+std::to_string(indexes[idx_iter])+" versus "+std::to_string(parameter_count));
-    if(indexes[idx_iter] >= parameters.size())Rcpp::stop("Index out of range (pars): case 2 indexes: "+std::to_string(indexes[idx_iter])+" versus "+std::to_string(parameters.size()));
+      if((unsigned)idx_iter >= indexes.size())Rcpp::stop("Index out of range: case 2 idx iter: "+std::to_string(idx_iter)+" versus "+std::to_string(indexes.size()));
+      if(indexes[idx_iter] >= parameter_count)Rcpp::stop("Index out of range: case 2 indexes: "+std::to_string(indexes[idx_iter])+" versus "+std::to_string(parameter_count));
+      if(indexes[idx_iter] >= parameters.size())Rcpp::stop("Index out of range (pars): case 2 indexes: "+std::to_string(indexes[idx_iter])+" versus "+std::to_string(parameters.size()));
 #endif
-    
-    stack.push(parameters[indexes[idx_iter]]);
-    if constexpr (dydx == CalcDyDx::BetaFirst || dydx == CalcDyDx::BetaSecond){
-      for(int idx = 0; idx < parameter_count; idx++){
-        if(idx == indexes[idx_iter]){
-          first_dx[idx].push(1.0);
-        } else {
-          first_dx[idx].push(0.0);
+      
+      stack.push(parameters[indexes[idx_iter]]);
+      if constexpr (dydx == CalcDyDx::BetaFirst || dydx == CalcDyDx::BetaSecond){
+        for(int idx = 0; idx < parameter_count; idx++){
+          if(idx == indexes[idx_iter]){
+            first_dx[idx].push(1.0);
+          } else {
+            first_dx[idx].push(0.0);
+          }
         }
       }
-    }
-    if constexpr (dydx == CalcDyDx::BetaSecond)addZeroDx2();
-    if constexpr (dydx == CalcDyDx::XBeta){
-      first_dx[0].push(0.0);
-      for(int idx = 0; idx < parameter_count; idx++){
-        if(idx == indexes[idx_iter]){
-          first_dx[idx+1].push(1.0);
-        } else {
-          first_dx[idx+1].push(0.0);
+      if constexpr (dydx == CalcDyDx::BetaSecond)addZeroDx2();
+      if constexpr (dydx == CalcDyDx::XBeta){
+        first_dx[0].push(0.0);
+        for(int idx = 0; idx < parameter_count; idx++){
+          if(idx == indexes[idx_iter]){
+            first_dx[idx+1].push(1.0);
+          } else {
+            first_dx[idx+1].push(0.0);
+          }
         }
+        addZeroDx2();
       }
-      addZeroDx2();
+      if constexpr (dydx == CalcDyDx::Zu)addZeroDx();
+      idx_iter++;
+      break;
     }
-    if constexpr (dydx == CalcDyDx::Zu)addZeroDx();
-    idx_iter++;
-    break;
-  }
     case Do::Add:
-  {
+    {
 #if defined(ENABLE_DEBUG) && defined(R_BUILD)
-    if(stack.size()<2)Rcpp::stop("Stack too small (3)");
+      if(stack.size()<2)Rcpp::stop("Stack too small (3)");
 #endif
-    a = stack.top();
-    stack.pop();
-    b = stack.top();
-    stack.pop();
-    stack.push(a+b);
-    if constexpr (dydx != CalcDyDx::None){
-      for(auto& fstack: first_dx){
-        a = fstack.top();
-        fstack.pop();
-        b = fstack.top();
-        fstack.pop();
-        fstack.push(a+b);
+      a = stack.top();
+      stack.pop();
+      b = stack.top();
+      stack.pop();
+      stack.push(a+b);
+      if constexpr (dydx != CalcDyDx::None){
+        for(auto& fstack: first_dx){
+          a = fstack.top();
+          fstack.pop();
+          b = fstack.top();
+          fstack.pop();
+          fstack.push(a+b);
+        }
       }
-    }
-    if constexpr (dydx == CalcDyDx::BetaSecond || dydx == CalcDyDx::XBeta){
-      for(auto& sstack: second_dx){
-        a = sstack.top();
-        sstack.pop();
-        b = sstack.top();
-        sstack.pop();
-        sstack.push(a+b);
+      if constexpr (dydx == CalcDyDx::BetaSecond || dydx == CalcDyDx::XBeta){
+        for(auto& sstack: second_dx){
+          a = sstack.top();
+          sstack.pop();
+          b = sstack.top();
+          sstack.pop();
+          sstack.push(a+b);
+        }
       }
+      break;
     }
-    break;
-  }
     case Do::Subtract:
-  {
+    {
 #if defined(ENABLE_DEBUG) && defined(R_BUILD)
-    if(stack.size()<2)Rcpp::stop("Stack too small (4)");
+      if(stack.size()<2)Rcpp::stop("Stack too small (4)");
 #endif
-    a = stack.top();
-    stack.pop();
-    b = stack.top();
-    stack.pop();
-    stack.push(a-b);
-    if constexpr (dydx != CalcDyDx::None){
-      for(auto& fstack: first_dx){
-        a = fstack.top();
-        fstack.pop();
-        b = fstack.top();
-        fstack.pop();
-        fstack.push(a-b);
+      a = stack.top();
+      stack.pop();
+      b = stack.top();
+      stack.pop();
+      stack.push(a-b);
+      if constexpr (dydx != CalcDyDx::None){
+        for(auto& fstack: first_dx){
+          a = fstack.top();
+          fstack.pop();
+          b = fstack.top();
+          fstack.pop();
+          fstack.push(a-b);
+        }
       }
-    }
-    if constexpr (dydx == CalcDyDx::BetaSecond || dydx == CalcDyDx::XBeta){
-      for(auto& sstack: second_dx){
-        a = sstack.top();
-        sstack.pop();
-        b = sstack.top();
-        sstack.pop();
-        sstack.push(a-b);
+      if constexpr (dydx == CalcDyDx::BetaSecond || dydx == CalcDyDx::XBeta){
+        for(auto& sstack: second_dx){
+          a = sstack.top();
+          sstack.pop();
+          b = sstack.top();
+          sstack.pop();
+          sstack.push(a-b);
+        }
       }
+      break;
     }
-    break;
-  }
     case Do::Multiply:
-  {
+    {
 #if defined(ENABLE_DEBUG) && defined(R_BUILD)
-    if(stack.size()<2)Rcpp::stop("Stack too small (5)");
+      if(stack.size()<2)Rcpp::stop("Stack too small (5)");
 #endif
-    
-    a = stack.top();
-    stack.pop();
-    b = stack.top();
-    stack.pop();
-    stack.push(a*b);
-    if constexpr (dydx != CalcDyDx::None){
-      dblvec a_top_dx;
-      dblvec b_top_dx;
-      for(auto& fstack: first_dx){
-        a_top_dx.push_back(fstack.top());
-        fstack.pop();
-        b_top_dx.push_back(fstack.top());
-        fstack.pop();
-        fstack.push(a*b_top_dx.back() + b*a_top_dx.back());
-      }
-      if constexpr (dydx == CalcDyDx::BetaSecond){
-        int index_count = 0;
-        for(int idx = 0; idx < parameter_count; idx++){
-          for(int jdx = idx; jdx < parameter_count; jdx++){
-            double adx2 = second_dx[index_count].top();
-            second_dx[index_count].pop();
-            double bdx2 = second_dx[index_count].top();
-            second_dx[index_count].pop();
-            double result = a*bdx2 + b*adx2 + a_top_dx[idx]*b_top_dx[jdx] + a_top_dx[jdx]*b_top_dx[idx];
-            second_dx[index_count].push(result);
-            index_count++;
+      
+      a = stack.top();
+      stack.pop();
+      b = stack.top();
+      stack.pop();
+      stack.push(a*b);
+      if constexpr (dydx != CalcDyDx::None){
+        dblvec a_top_dx;
+        dblvec b_top_dx;
+        for(auto& fstack: first_dx){
+          a_top_dx.push_back(fstack.top());
+          fstack.pop();
+          b_top_dx.push_back(fstack.top());
+          fstack.pop();
+          fstack.push(a*b_top_dx.back() + b*a_top_dx.back());
+        }
+        if constexpr (dydx == CalcDyDx::BetaSecond){
+          int index_count = 0;
+          for(int idx = 0; idx < parameter_count; idx++){
+            for(int jdx = idx; jdx < parameter_count; jdx++){
+              double adx2 = second_dx[index_count].top();
+              second_dx[index_count].pop();
+              double bdx2 = second_dx[index_count].top();
+              second_dx[index_count].pop();
+              double result = a*bdx2 + b*adx2 + a_top_dx[idx]*b_top_dx[jdx] + a_top_dx[jdx]*b_top_dx[idx];
+              second_dx[index_count].push(result);
+              index_count++;
+            }
+          }
+        }
+        if constexpr (dydx == CalcDyDx::XBeta){
+          for(int jdx = 0; jdx < parameter_count; jdx++){
+            double adx2 = second_dx[jdx].top();
+            second_dx[jdx].pop();
+            double bdx2 = second_dx[jdx].top();
+            second_dx[jdx].pop();
+            double result = a*bdx2 + b*adx2 + a_top_dx[0]*b_top_dx[jdx+1] + a_top_dx[jdx+1]*b_top_dx[0];
+            second_dx[jdx].push(result);
           }
         }
       }
-      if constexpr (dydx == CalcDyDx::XBeta){
-        for(int jdx = 0; jdx < parameter_count; jdx++){
-          double adx2 = second_dx[jdx].top();
-          second_dx[jdx].pop();
-          double bdx2 = second_dx[jdx].top();
-          second_dx[jdx].pop();
-          double result = a*bdx2 + b*adx2 + a_top_dx[0]*b_top_dx[jdx+1] + a_top_dx[jdx+1]*b_top_dx[0];
-          second_dx[jdx].push(result);
-        }
-      }
+      break;
     }
-    break;
-  }
     case Do::Divide:
-  {
+    {
 #if defined(ENABLE_DEBUG) && defined(R_BUILD)
-    if(stack.size()<2)Rcpp::stop("Stack too small (6)");
+      if(stack.size()<2)Rcpp::stop("Stack too small (6)");
 #endif
-    
-    a = stack.top();
-    stack.pop();
-    b = stack.top();
-    
+      
+      a = stack.top();
+      stack.pop();
+      b = stack.top();
+      
 #if defined(ENABLE_DEBUG) && defined(R_BUILD)
-    if(b == 0)Rcpp::stop("Divide by zero (6)");
+      if(b == 0.0)Rcpp::stop("Divide by zero (6)");
+      if(isinf(a/b))Rcpp::stop("Infinity in division (6b)");
 #endif
-    
-    stack.pop();
-    stack.push(a/b);
-    if constexpr (dydx != CalcDyDx::None){
-      dblvec a_top_dx;
-      dblvec b_top_dx;
-      for(auto& fstack: first_dx){
-        a_top_dx.push_back(fstack.top());
-        fstack.pop();
-        b_top_dx.push_back(fstack.top());
-        fstack.pop();
-        double result = (b*a_top_dx.back() - a*b_top_dx.back())/(b*b);
-        fstack.push(result);
-      }
-      if constexpr (dydx == CalcDyDx::BetaSecond){
-        int index_count = 0;
-        for(int idx = 0; idx < parameter_count; idx++){
-          for(int jdx = idx; jdx < parameter_count; jdx++){
-            double adx2 = second_dx[index_count].top();
-            second_dx[index_count].pop();
-            double bdx2 = second_dx[index_count].top();
-            second_dx[index_count].pop();
-            double result = (adx2*b - a_top_dx[idx]*b_top_dx[jdx]- a_top_dx[jdx]*b_top_dx[idx])/(b*b) - (a*bdx2*b - 2*b_top_dx[idx]*b_top_dx[jdx])/(b*b*b);
-            second_dx[index_count].push(result);
-            index_count++;
+      
+      stack.pop();
+      stack.push(a/b);
+      if constexpr (dydx != CalcDyDx::None){
+        dblvec a_top_dx;
+        dblvec b_top_dx;
+        for(auto& fstack: first_dx){
+          a_top_dx.push_back(fstack.top());
+          fstack.pop();
+          b_top_dx.push_back(fstack.top());
+          fstack.pop();
+          double result = (b*a_top_dx.back() - a*b_top_dx.back())/(b*b);
+          fstack.push(result);
+        }
+        if constexpr (dydx == CalcDyDx::BetaSecond){
+          int index_count = 0;
+          for(int idx = 0; idx < parameter_count; idx++){
+            for(int jdx = idx; jdx < parameter_count; jdx++){
+              double adx2 = second_dx[index_count].top();
+              second_dx[index_count].pop();
+              double bdx2 = second_dx[index_count].top();
+              second_dx[index_count].pop();
+              double result = (adx2*b - a_top_dx[idx]*b_top_dx[jdx]- a_top_dx[jdx]*b_top_dx[idx])/(b*b) - (a*bdx2*b - 2*b_top_dx[idx]*b_top_dx[jdx])/(b*b*b);
+              second_dx[index_count].push(result);
+              index_count++;
+            }
+          }
+        }
+        if constexpr (dydx == CalcDyDx::XBeta){
+          for(int jdx = 0; jdx < parameter_count; jdx++){
+            double adx2 = second_dx[jdx].top();
+            second_dx[jdx].pop();
+            double bdx2 = second_dx[jdx].top();
+            second_dx[jdx].pop();
+            double result = (adx2*b - a_top_dx[0]*b_top_dx[jdx+1]- a_top_dx[jdx+1]*b_top_dx[0])/(b*b) - (a*bdx2*b - 2*b_top_dx[0]*b_top_dx[jdx])/(b*b*b);
+            second_dx[jdx].push(result);
           }
         }
       }
-      if constexpr (dydx == CalcDyDx::XBeta){
-        for(int jdx = 0; jdx < parameter_count; jdx++){
-          double adx2 = second_dx[jdx].top();
-          second_dx[jdx].pop();
-          double bdx2 = second_dx[jdx].top();
-          second_dx[jdx].pop();
-          double result = (adx2*b - a_top_dx[0]*b_top_dx[jdx+1]- a_top_dx[jdx+1]*b_top_dx[0])/(b*b) - (a*bdx2*b - 2*b_top_dx[0]*b_top_dx[jdx])/(b*b*b);
-          second_dx[jdx].push(result);
-        }
-      }
+      break;
     }
-    break;
-  }
     case Do::Sqrt:
       
 #if defined(ENABLE_DEBUG) && defined(R_BUILD)
@@ -537,70 +549,70 @@ inline dblvec glmmr::calculator::calculate(const int i,
       }
       break;
     case Do::Power:
-      {
+    {
 #if defined(ENABLE_DEBUG) && defined(R_BUILD)
-        if(stack.size()<2)Rcpp::stop("Stack too small (8)");
+      if(stack.size()<2)Rcpp::stop("Stack too small (8)");
 #endif
-        
-        a = stack.top();
-        stack.pop();
-        b = stack.top();
-        stack.pop();
-        double out = pow(a,b);
-        
+      
+      a = stack.top();
+      stack.pop();
+      b = stack.top();
+      stack.pop();
+      double out = pow(a,b);
+      
 #ifdef R_BUILD
-        // try and catch these possible failures as it seems to cause crash if nan allowed to propogate
-        if(out != out)Rcpp::stop("Exponent fail: "+std::to_string(a)+"^"+std::to_string(b));
+      // try and catch these possible failures as it seems to cause crash if nan allowed to propogate
+      if(out != out)Rcpp::stop("Exponent fail: "+std::to_string(a)+"^"+std::to_string(b));
 #endif
-        
-        stack.push(out);
-        
-        if constexpr (dydx != CalcDyDx::None){
-          dblvec a_top_dx;
-          dblvec b_top_dx;
-          for(auto& fstack: first_dx){
-            a_top_dx.push_back(fstack.top());
-            fstack.pop();
-            b_top_dx.push_back(fstack.top());
-            fstack.pop();
-            double result = pow(a,b)*b_top_dx.back()*log(a) + pow(a,b-1)*b*a_top_dx.back();
+      
+      stack.push(out);
+      
+      if constexpr (dydx != CalcDyDx::None){
+        dblvec a_top_dx;
+        dblvec b_top_dx;
+        for(auto& fstack: first_dx){
+          a_top_dx.push_back(fstack.top());
+          fstack.pop();
+          b_top_dx.push_back(fstack.top());
+          fstack.pop();
+          double result = pow(a,b)*b_top_dx.back()*log(a) + pow(a,b-1)*b*a_top_dx.back();
 #ifdef R_BUILD
-            // this can sometimes result in a crash if the values of the parameters aren't right
-            if(result != result)Rcpp::stop("Exponent dydx fail: "+std::to_string(a)+"^"+std::to_string(b-1));
+          // this can sometimes result in a crash if the values of the parameters aren't right
+          if(result != result)Rcpp::stop("Exponent dydx fail: "+std::to_string(a)+"^"+std::to_string(b-1));
 #endif
-            fstack.push(result);
-          }
-          if constexpr (dydx == CalcDyDx::BetaSecond){
-            int index_count = 0;
-            for(int idx = 0; idx < parameter_count; idx++){
-              for(int jdx = idx; jdx < parameter_count; jdx++){
-                double adx2 = second_dx[index_count].top();
-                second_dx[index_count].pop();
-                double bdx2 = second_dx[index_count].top();
-                second_dx[index_count].pop();
-                double result1 = first_dx[jdx].top()*b_top_dx[idx]*log(a) + stack.top()*(bdx2*log(a) + b_top_dx[idx]*(1/a)*a_top_dx[jdx]);
-                double result2 = pow(a,b-1)*b_top_dx[jdx]*log(a) + pow(a,b-2)*(b-1)*a_top_dx[jdx];
-                double result3 = result2*b*a_top_dx[idx] + pow(a,b-1)*(b*adx2+b_top_dx[jdx]*a_top_dx[idx]);
-                second_dx[index_count].push(result1 + result3);
-                index_count++;
-              }
-            }
-          }
-          if constexpr (dydx == CalcDyDx::XBeta){
-            for(int jdx = 0; jdx < parameter_count; jdx++){
-              double adx2 = second_dx[jdx].top();
-              second_dx[jdx].pop();
-              double bdx2 = second_dx[jdx].top();
-              second_dx[jdx].pop();
-              double result1 = first_dx[jdx+1].top()*b_top_dx[0]*log(a) + stack.top()*(bdx2*log(a) + b_top_dx[0]*(1/a)*a_top_dx[jdx+1]);
-              double result2 = pow(a,b-1)*b_top_dx[jdx+1]*log(a) + pow(a,b-2)*(b-1)*a_top_dx[jdx+1];
-              double result3 = result2*b*a_top_dx[0] + pow(a,b-1)*(b*adx2+b_top_dx[jdx+1]*a_top_dx[0]);
-              second_dx[jdx].push(result1 + result3);
+          fstack.push(result);
+        }
+        if constexpr (dydx == CalcDyDx::BetaSecond){
+          int index_count = 0;
+          for(int idx = 0; idx < parameter_count; idx++){
+            for(int jdx = idx; jdx < parameter_count; jdx++){
+              double adx2 = second_dx[index_count].top();
+              second_dx[index_count].pop();
+              double bdx2 = second_dx[index_count].top();
+              second_dx[index_count].pop();
+              double result1 = first_dx[jdx].top()*b_top_dx[idx]*log(a) + stack.top()*(bdx2*log(a) + b_top_dx[idx]*(1/a)*a_top_dx[jdx]);
+              double result2 = pow(a,b-1)*b_top_dx[jdx]*log(a) + pow(a,b-2)*(b-1)*a_top_dx[jdx];
+              double result3 = result2*b*a_top_dx[idx] + pow(a,b-1)*(b*adx2+b_top_dx[jdx]*a_top_dx[idx]);
+              second_dx[index_count].push(result1 + result3);
+              index_count++;
             }
           }
         }
-        break;
+        if constexpr (dydx == CalcDyDx::XBeta){
+          for(int jdx = 0; jdx < parameter_count; jdx++){
+            double adx2 = second_dx[jdx].top();
+            second_dx[jdx].pop();
+            double bdx2 = second_dx[jdx].top();
+            second_dx[jdx].pop();
+            double result1 = first_dx[jdx+1].top()*b_top_dx[0]*log(a) + stack.top()*(bdx2*log(a) + b_top_dx[0]*(1/a)*a_top_dx[jdx+1]);
+            double result2 = pow(a,b-1)*b_top_dx[jdx+1]*log(a) + pow(a,b-2)*(b-1)*a_top_dx[jdx+1];
+            double result3 = result2*b*a_top_dx[0] + pow(a,b-1)*(b*adx2+b_top_dx[jdx+1]*a_top_dx[0]);
+            second_dx[jdx].push(result1 + result3);
+          }
+        }
       }
+      break;
+    }
     case Do::Exp:
       
 #if defined(ENABLE_DEBUG) && defined(R_BUILD)
@@ -959,17 +971,17 @@ inline dblvec glmmr::calculator::calculate(const int i,
       }
       break;
     case Do::PushExtraData:
-      {
-        stack.push(extraData);
-        allDyDxZero();
-        break;
-      }
+    {
+      stack.push(extraData);
+      allDyDxZero();
+      break;
+    }
     case Do::PushY:
-      {
-        stack.push(y[i]);
-        allDyDxZero();
-        break;
-      }
+    {
+      stack.push(y[i]);
+      allDyDxZero();
+      break;
+    }
     case Do::Int10:
       stack.push(10);
       allDyDxZero();
@@ -1039,38 +1051,38 @@ inline dblvec glmmr::calculator::calculate(const int i,
       allDyDxZero();
       break;
     case Do::LogFactorialApprox:
-      {
-        //log factorial approximation
-        #if defined(ENABLE_DEBUG) && defined(R_BUILD)
-        if(stack.size()==0)Rcpp::stop("Stack too small (40)");
-        #endif
-        
-        a = stack.top();
-        stack.pop();
-        // Ramanujan approximation
-        if(a == 0){
-          stack.push(0);
-        } else {
-          double result = a*log(a) - a + log(a*(1+4*a*(1+2*a)))/6 + log(3.141593)/2;
-          stack.push(result);
-        }
-        // NOTE: this function is only ever used in Poisson/binom log likelihood and so the top of the derivative
-        // stacks should be 0, so we don't need to do anything. However, this should be updated if ever this
-        // function is used more broadly.
-        //if(order > 0){
-        //  addZeroDx();
-        //}
-        //if(order == 2){
-        //  addZeroDx2();
-        //}
-        break;
+    {
+      //log factorial approximation
+#if defined(ENABLE_DEBUG) && defined(R_BUILD)
+      if(stack.size()==0)Rcpp::stop("Stack too small (40)");
+#endif
+      
+      a = stack.top();
+      stack.pop();
+      // Ramanujan approximation
+      if(a == 0){
+        stack.push(0);
+      } else {
+        double result = a*log(a) - a + log(a*(1+4*a*(1+2*a)))/6 + log(3.141593)/2;
+        stack.push(result);
       }
+      // NOTE: this function is only ever used in Poisson/binom log likelihood and so the top of the derivative
+      // stacks should be 0, so we don't need to do anything. However, this should be updated if ever this
+      // function is used more broadly.
+      //if(order > 0){
+      //  addZeroDx();
+      //}
+      //if(order == 2){
+      //  addZeroDx2();
+      //}
+      break;
+    }
     case Do::PushVariance:
-      {
-        stack.push(variance(i));
-        allDyDxZero();
-        break;
-      }
+    {
+      stack.push(variance(i));
+      allDyDxZero();
+      break;
+    }
     case Do::PushUserNumber0:
       stack.push(numbers[0]);
       allDyDxZero();
@@ -1111,12 +1123,52 @@ inline dblvec glmmr::calculator::calculate(const int i,
       stack.push(numbers[9]);
       allDyDxZero();
       break;
+    case Do::PushUserNumber10:
+      stack.push(numbers[10]);
+      allDyDxZero();
+      break;
+    case Do::PushUserNumber11:
+      stack.push(numbers[11]);
+      allDyDxZero();
+      break;
+    case Do::PushUserNumber12:
+      stack.push(numbers[12]);
+      allDyDxZero();
+      break;
+    case Do::PushUserNumber13:
+      stack.push(numbers[13]);
+      allDyDxZero();
+      break;
+    case Do::PushUserNumber14:
+      stack.push(numbers[14]);
+      allDyDxZero();
+      break;
+    case Do::PushUserNumber15:
+      stack.push(numbers[15]);
+      allDyDxZero();
+      break;
+    case Do::PushUserNumber16:
+      stack.push(numbers[16]);
+      allDyDxZero();
+      break;
+    case Do::PushUserNumber17:
+      stack.push(numbers[17]);
+      allDyDxZero();
+      break;
+    case Do::PushUserNumber18:
+      stack.push(numbers[18]);
+      allDyDxZero();
+      break;
+    case Do::PushUserNumber19:
+      stack.push(numbers[19]);
+      allDyDxZero();
+      break;
     }
     
-    #if defined(ENABLE_DEBUG) && defined(R_BUILD)
+#if defined(ENABLE_DEBUG) && defined(R_BUILD)
     if(stack.size() == 0)Rcpp::stop("Error stack empty");
     if(stack.top() != stack.top() || isnan(stack.top()))Rcpp::stop("Calculation evaluates to NaN");
-    #endif
+#endif
   }
   
 #if defined(ENABLE_DEBUG) && defined(R_BUILD)
@@ -1128,9 +1180,9 @@ inline dblvec glmmr::calculator::calculate(const int i,
   
   if constexpr (dydx != CalcDyDx::None){
     for(const auto& fstack: first_dx){
-      #if defined(ENABLE_DEBUG) && defined(R_BUILD)
+#if defined(ENABLE_DEBUG) && defined(R_BUILD)
       if(fstack.size()==0)Rcpp::stop("Error derivative stack empty");
-      #endif
+#endif
       result.push_back(fstack.top());
     }
   }
@@ -1148,16 +1200,16 @@ inline dblvec glmmr::calculator::calculate(const int i,
 
 inline void glmmr::calculator::print_instructions() const {
   //currently only setup for R
-  #ifdef R_BUILD
+#ifdef R_BUILD
   int counter = 1;
   int idx_iter = 0;
   Rcpp::Rcout << "\nInstructions:\n";
   for(const auto& i: instructions){
     Rcpp::Rcout << counter << ". " << instruction_str.at(i);
     switch(i){
-      case Do::PushUserNumber0:
-        Rcpp::Rcout << " = " << numbers[0] << "\n";
-        break;
+    case Do::PushUserNumber0:
+      Rcpp::Rcout << " = " << numbers[0] << "\n";
+      break;
     case Do::PushUserNumber1:
       Rcpp::Rcout << " = " << numbers[1] << "\n";
       break;
@@ -1202,11 +1254,11 @@ inline void glmmr::calculator::print_instructions() const {
     }
     counter++;
   }
-  #endif
+#endif
 }
 
 inline void glmmr::calculator::print_names(bool data, bool parameters) const{
-  #ifdef R_BUILD
+#ifdef R_BUILD
   if(data){
     Rcpp::Rcout << "\nData names: \n";
     glmmr::print_vec_1d<strvec>(data_names);
@@ -1215,5 +1267,5 @@ inline void glmmr::calculator::print_names(bool data, bool parameters) const{
     Rcpp::Rcout << "\nParameter names: \n";
     glmmr::print_vec_1d<strvec>(parameter_names);
   }
-  #endif
+#endif
 }
